/*
 * $XConsortium: SSMailbox.c,v 1.35 89/10/09 16:51:44 jim Exp $
 *
 * Copyright 1988 Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of M.I.T. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  M.I.T. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * Author:  Jim Fulton, MIT X Consortium
 *
 * I recommend that you use the new mailfull and mailempty bitmaps instead of
 * the ugly mailboxes:
 *
 *         XBiff*fullPixmap:  mailfull
 *         XBiff*emptyPixmap:  mailempty
 */

#include <stdio.h>			/* for printing error messages */
#include <pwd.h>			/* for getting username */

#include <X11/cursorfont.h>		/* for cursor constants */
#include <X11/StringDefs.h>		/* for useful atom names */
#include <X11/IntrinsicP.h>		/* for toolkit stuff */
#include <sys/stat.h>			/* for stat() ** needs types.h ***/

#include <X11/bitmaps/mailfull>		/* for flag up (mail present) bits */
#include <X11/bitmaps/mailempty>	/* for flag down (mail not here) */

#include <X11/Xaw/XawInit.h>
#include "SSMailboxP.h"			/* for implementation mailbox stuff */

#include <X11/Xmu/Converters.h>		/* for XmuCvtStringToBitmap */

#ifdef SHAPE
#include <X11/extensions/shape.h>
#endif

#include "strtok.h"

/*
 * The default user interface is to have the mailbox turn itself off whenever
 * the user presses a button in it.  Expert users might want to make this 
 * happen on EnterWindow.  It might be nice to provide support for some sort of
 * exit callback so that you can do things like press q to quit.
 */

static char defaultTranslations[] = 
  "<Btn1Down>:  unset()\n\
   <Btn3Down>:  review()";

static void Check(), Set(), Unset(), Review();
static Boolean playOneWP(), playManyWP();

static XtActionsRec actionsList[] = { 
    { "check",	Check },
    { "unset",	Unset },
    { "set",	Set },
    { "review", Review },
};


/* Initialization of defaults */

#define offset(field) XtOffset(SSMailboxWidget,ssmailbox.field)
#define goffset(field) XtOffset(Widget,core.field)

static Dimension defDim = 48;
static Pixmap nopix = None;

static XtResource resources[] = {
    { XtNwidth, XtCWidth, XtRDimension, sizeof (Dimension), 
	goffset (width), XtRDimension, (caddr_t)&defDim },
    { XtNheight, XtCHeight, XtRDimension, sizeof (Dimension),
	goffset (height), XtRDimension, (caddr_t)&defDim },
    { XtNupdate, XtCInterval, XtRInt, sizeof (int),
	offset (update), XtRString, "30" },
    { XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	offset (foreground_pixel), XtRString, "black" },
    { XtNbackground, XtCBackground, XtRPixel, sizeof (Pixel),
	goffset (background_pixel), XtRString, "white" },
    { XtNreverseVideo, XtCReverseVideo, XtRBoolean, sizeof (Boolean),
	offset (reverseVideo), XtRString, "FALSE" },
    { XtNfile, XtCFile, XtRString, sizeof (String),
	offset (filename), XtRString, NULL },
    { XtNcheckCommand, XtCCheckCommand, XtRString, sizeof(char*),
	offset (check_command), XtRString, NULL},
    { XtNvolume, XtCVolume, XtRInt, sizeof(int),
	offset (volume), XtRString, "50"},
    { XtNonceOnly, XtCBoolean, XtRBoolean, sizeof(Boolean),
	offset (once_only), XtRImmediate, (caddr_t)False },
    { XtNfullPixmap, XtCPixmap, XtRBitmap, sizeof(Pixmap),
	offset (full.bitmap), XtRString, "flagup" },
    { XtNfullPixmapMask, XtCPixmapMask, XtRBitmap, sizeof(Pixmap),
	offset (full.mask), XtRBitmap, (caddr_t) &nopix },
    { XtNemptyPixmap, XtCPixmap, XtRBitmap, sizeof(Pixmap),
	offset (empty.bitmap), XtRString, "flagdown" },
    { XtNemptyPixmapMask, XtCPixmapMask, XtRBitmap, sizeof(Pixmap),
	offset (empty.mask), XtRBitmap, (caddr_t) &nopix },
    { XtNflip, XtCFlip, XtRBoolean, sizeof(Boolean),
	offset (flipit), XtRString, "true" },
#ifdef SHAPE
    { XtNshapeWindow, XtCShapeWindow, XtRBoolean, sizeof(Boolean),
        offset (shapeit), XtRString, "false" },
#endif
    { XtNsound, XtCSound, XtRString, sizeof (String),
	offset (sound), XtRString, NULL },
    { XtNsoundBindings, XtCSoundBindings, XtRString, sizeof (String),
	offset (sound_bindings), XtRString, NULL },
};

#undef offset
#undef goffset

static void GetMailFile(), CloseDown();
static void check_ssmailbox(), redraw_ssmailbox(), beep();
static void ClassInitialize(), Initialize(), Realize(), Destroy(), Redisplay();
static Boolean SetValues();

SSMailboxClassRec ssmailboxClassRec = {
    { /* core fields */
    /* superclass		*/	&widgetClassRec,
    /* class_name		*/	"SSMailbox",
    /* widget_size		*/	sizeof(SSMailboxRec),
    /* class_initialize		*/	ClassInitialize,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	Realize,
    /* actions			*/	actionsList,
    /* num_actions		*/	XtNumber(actionsList),
    /* resources		*/	resources,
    /* resource_count		*/	XtNumber(resources),
    /* xrm_class		*/	NULL,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	Destroy,
    /* resize			*/	NULL,
    /* expose			*/	Redisplay,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	defaultTranslations,
    /* query_geometry		*/	XtInheritQueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
    }
};

struct SoundStruct 
{
    char *sound;
    struct SoundStruct *next;
};

struct MsgStruct 
{
    struct SoundStruct *sounds;
    struct MsgStruct *next;
};

struct PlaySound
{
    Display *dpy;
    int volume;
    int num_sounds;
    struct MsgStruct *msg;
    String sound;
};

WidgetClass ssmailboxWidgetClass = (WidgetClass) &ssmailboxClassRec;

/*
 * widget initialization
 */

static void ClassInitialize ()
{
    static XtConvertArgRec screenConvertArg[] = {
    { XtWidgetBaseOffset, (caddr_t) XtOffset(Widget, core.screen), sizeof(Screen *) }
    };

    XawInitializeWidgetSet();
    XtAddConverter (XtRString, XtRBitmap, XmuCvtStringToBitmap,
		    screenConvertArg, XtNumber(screenConvertArg));
    return;
}

static GC get_ssmailbox_gc (w)
    SSMailboxWidget w;
{
    XtGCMask valuemask;
    XGCValues xgcv;

    valuemask = GCForeground | GCBackground | GCFunction | GCGraphicsExposures;
    xgcv.foreground = w->ssmailbox.foreground_pixel;
    xgcv.background = w->core.background_pixel;
    xgcv.function = GXcopy;
    xgcv.graphics_exposures = False;	/* this is Bool, not Boolean */
    return (XtGetGC ((Widget) w, valuemask, &xgcv));
}


/* ARGSUSED */
static void Initialize (request, new)
    Widget request, new;
{
    SSMailboxWidget w = (SSMailboxWidget) new;
#ifdef SHAPE
    int shape_event_base, shape_error_base;
#endif

    if (!w->ssmailbox.filename) GetMailFile (w);

    if (w->core.width <= 0) w->core.width = 1;
    if (w->core.height <= 0) w->core.height = 1;

    if (w->ssmailbox.reverseVideo) {
	Pixel tmp;

	tmp = w->ssmailbox.foreground_pixel;
	w->ssmailbox.foreground_pixel = w->core.background_pixel;
	w->core.background_pixel = tmp;
    }

#ifdef SHAPE
    if (w->ssmailbox.shapeit && !XShapeQueryExtension (XtDisplay (w),
						     &shape_event_base,
						     &shape_error_base))
      w->ssmailbox.shapeit = False;
    w->ssmailbox.shape_cache.mask = None;
#endif

    w->ssmailbox.gc = get_ssmailbox_gc (w);
    w->ssmailbox.interval_id = (XtIntervalId) 0;
    w->ssmailbox.full.pixmap = None;
    w->ssmailbox.empty.pixmap = None;
    w->ssmailbox.last_numnew = 0;
    w->ssmailbox.bindings = 0;

    return;
}


/*
 * action procedures
 */

/*
 * pretend there is new mail; put widget in flagup state
 */

/* ARGSUSED */
static void Set (gw, event, params, nparams)
    Widget gw;
    XEvent *event;
    String *params;
    Cardinal *nparams;
{
    SSMailboxWidget w = (SSMailboxWidget) gw;

    w->ssmailbox.last_numnew = -1;

    check_ssmailbox (w, TRUE, FALSE);	/* redraw, no reset */

    return;
}


/*
 * ack the existing mail; put widget in flagdown state
 */

/* ARGSUSED */
static void Unset (gw, event, params, nparams)
    Widget gw;
    XEvent *event;
    String *params;
    Cardinal *nparams;
{
    SSMailboxWidget w = (SSMailboxWidget) gw;

    check_ssmailbox (w, TRUE, TRUE);	/* redraw, reset */

    return;
}

/*
 * pretend all mail is new; review sounds
 */

/* ARGSUSED */
static void Review (gw, event, params, nparams)
    Widget gw;
    XEvent *event;
    String *params;
    Cardinal *nparams;
{
    SSMailboxWidget w = (SSMailboxWidget) gw;

    w->ssmailbox.last_numnew = 0;
    check_ssmailbox (w, TRUE, FALSE);	/* redraw, reset */

    return;
}


/*
 * look to see if there is new mail; if so, Set, else Unset
 */

/* ARGSUSED */
static void Check (gw, event, params, nparams)
    Widget gw;
    XEvent *event;
    String *params;
    Cardinal *nparams;
{
    SSMailboxWidget w = (SSMailboxWidget) gw;

    check_ssmailbox (w, TRUE, FALSE);	/* redraw, no reset */

    return;
}


/* ARGSUSED */
static void clock_tic (client_data, id)
    caddr_t client_data;
    XtIntervalId *id;
{
    SSMailboxWidget w = (SSMailboxWidget) client_data;

    check_ssmailbox (w, FALSE, FALSE);	/* no redraw, no reset */

    /*
     * and reset the timer
     */

    w->ssmailbox.interval_id = XtAddTimeOut (w->ssmailbox.update * 1000,
					   clock_tic, (caddr_t) w);

    return;
}

static Pixmap make_pixmap (dpy, w, bitmap, depth, flip, widthp, heightp)
    Display *dpy;
    SSMailboxWidget w;
    Pixmap bitmap;
    Boolean flip;
    int depth;
    int *widthp, *heightp;
{
    Window root;
    int x, y;
    unsigned int width, height, bw, dep;
    unsigned long fore, back;

    if (!XGetGeometry (dpy, bitmap, &root, &x, &y, &width, &height, &bw, &dep))
      return None;

    *widthp = (int) width;
    *heightp = (int) height;
    if (flip) {
	fore = w->core.background_pixel;
	back = w->ssmailbox.foreground_pixel;
    } else {
	fore = w->ssmailbox.foreground_pixel;
	back = w->core.background_pixel;
    }
    return XmuCreatePixmapFromBitmap (dpy, w->core.window, bitmap, 
				      width, height, depth, fore, back);
}

static void Realize (gw, valuemaskp, attr)
    Widget gw;
    XtValueMask *valuemaskp;
    XSetWindowAttributes *attr;
{
    SSMailboxWidget w = (SSMailboxWidget) gw;
    register Display *dpy = XtDisplay (w);
    int depth = w->core.depth;

    *valuemaskp |= (CWBitGravity | CWCursor);
    attr->bit_gravity = ForgetGravity;
    attr->cursor = XCreateFontCursor (dpy, XC_top_left_arrow);

    XtCreateWindow (gw, InputOutput, (Visual *) CopyFromParent,
		    *valuemaskp, attr);

    /*
     * build up the pixmaps that we'll put into the image
     */
    if (w->ssmailbox.full.bitmap == None) {
	w->ssmailbox.full.bitmap = 
	  XCreateBitmapFromData (dpy, w->core.window, mailfull_bits,
				 mailfull_width, mailfull_height);
    }
    if (w->ssmailbox.empty.bitmap == None) {
	w->ssmailbox.empty.bitmap =
	  XCreateBitmapFromData (dpy, w->core.window, mailempty_bits,
				 mailempty_width, mailempty_height);
    }

    w->ssmailbox.empty.pixmap = make_pixmap (dpy, w, w->ssmailbox.empty.bitmap,
					   depth, False,
					   &w->ssmailbox.empty.width,
					   &w->ssmailbox.empty.height);
    w->ssmailbox.full.pixmap = make_pixmap (dpy, w, w->ssmailbox.full.bitmap,
					  depth, w->ssmailbox.flipit,
					  &w->ssmailbox.full.width,
					  &w->ssmailbox.full.height);
			 
#ifdef SHAPE
    if (w->ssmailbox.empty.mask == None && w->ssmailbox.full.mask == None)
      w->ssmailbox.shapeit = False;
#endif

    w->ssmailbox.interval_id = XtAddTimeOut (w->ssmailbox.update * 1000,
					   clock_tic, (caddr_t) w);

#ifdef SHAPE
    w->ssmailbox.shape_cache.mask = None;
#endif

    if (w->ssmailbox.sound_bindings) 
    {
	Bindings *binding;
	char *pair, *name, *sound;
	
	while (pair=strtoken(&w->ssmailbox.sound_bindings,",",1)) 
	{
            name=strtoken(&pair, "|", 1);
	    if (name) sound=strtoken(&pair, "|", 1);
	    if (!name || !sound)
	    {
		fprintf(stderr, "malformed sound bindings.\n");
		continue;
	    }
	    binding = (Bindings *)malloc(sizeof(Bindings));
	    binding->sound = sound;
	    if (binding->pattern = regcomp(name)) 
	    {
		binding->next = w->ssmailbox.bindings;
		w->ssmailbox.bindings = binding;
	    }
	}
    }

    return;
}


static void Destroy (gw)
    Widget gw;
{
    SSMailboxWidget w = (SSMailboxWidget) gw;
    Display *dpy = XtDisplay (gw);

    XtFree (w->ssmailbox.filename);
    if (w->ssmailbox.interval_id) XtRemoveTimeOut (w->ssmailbox.interval_id);
    XtDestroyGC (w->ssmailbox.gc);
#define freepix(p) if (p) XFreePixmap (dpy, p)
    freepix (w->ssmailbox.full.bitmap);		/* until cvter does ref cnt */
    freepix (w->ssmailbox.full.mask);		/* until cvter does ref cnt */
    freepix (w->ssmailbox.full.pixmap);
    freepix (w->ssmailbox.empty.bitmap);		/* until cvter does ref cnt */
    freepix (w->ssmailbox.empty.mask);		/* until cvter does ref cnt */
    freepix (w->ssmailbox.empty.pixmap);
#ifdef SHAPE
    freepix (w->ssmailbox.shape_cache.mask);
#endif
#undef freepix

    while (w->ssmailbox.bindings) {
	Bindings *b;
	XtFree(w->ssmailbox.bindings->pattern);
	b = w->ssmailbox.bindings->next;
	XtFree(w->ssmailbox.bindings);
	w->ssmailbox.bindings = b;
    }
    return;
}


static void Redisplay (gw)
    Widget gw;
{
    SSMailboxWidget w = (SSMailboxWidget) gw;

    check_ssmailbox (w, TRUE, FALSE);
}

static void free_sounds(sounds)
struct MsgStruct *sounds;
{
    struct MsgStruct *m;
    struct SoundStruct *s;

    while (sounds) {
	while (sounds->sounds) 
	{
	    s = sounds->sounds->next;
	    free((char *)sounds->sounds);
	    sounds->sounds = s;
	}
	m = sounds->next;
	free((char *)sounds);
	sounds = m;
    }
}

static void check_ssmailbox (w, force_redraw, reset)
    SSMailboxWidget w;
    Boolean force_redraw, reset;
{
    long numnew = 0;
    int beepit = 0;
    struct MsgStruct *sounds = 0;

    if (w->ssmailbox.check_command != NULL) {
	numnew = system(w->ssmailbox.check_command);
    }
    else {
	numnew = newmsgs(w, &sounds);
    }

    /*
     * Now check for changes.  If reset is set then we want to pretent that
     * there is no mail.  If the mailbox is empty then we want to turn off
     * the flag.  Otherwise if the mailbox has changed size then we want to
     * put the flag up.
     *
     * The cases are:
     *    o  forced reset by user                        DOWN
     *    o  no mailbox or empty (zero-sized) mailbox    DOWN
     *    o  same size as last time                      no change
     *    o  bigger than last time                       UP
     *    o  smaller than last time but non-zero         UP
     *
     * The last two cases can be expressed as different from last
     * time and non-zero.
     */

    if (reset) {			/* forced reset */
	w->ssmailbox.flag_up = FALSE;
	force_redraw = TRUE;
    } else if (numnew == 0) {
	w->ssmailbox.flag_up = FALSE;
	if (w->ssmailbox.last_numnew > 0) force_redraw = TRUE;  /* if change */
    } else if (numnew <= w->ssmailbox.last_numnew) {	/* less mail than
							   before */
	/* do nothing; don't play sounds and don't change icon since
	   user may have lowered the flag manually. */
    } else if (numnew > w->ssmailbox.last_numnew) {  /* different size */
	if (!w->ssmailbox.once_only || !w->ssmailbox.flag_up)
	    beepit = 1; 
	w->ssmailbox.flag_up = TRUE;
	force_redraw = TRUE;
    } 

    if (force_redraw)
	redraw_ssmailbox (w);

    if (beepit)
	beep(w, sounds, numnew - w->ssmailbox.last_numnew);
    else
	free_sounds(sounds);

    w->ssmailbox.last_numnew = numnew;

    return;
}

/*
 * get user name for building mailbox
 */

static void GetMailFile (w)
    SSMailboxWidget w;
{
    char *getlogin();
    char *username;

    username = getlogin ();
    if (!username) {
	struct passwd *pw = getpwuid (getuid ());

	if (!pw) {
	    fprintf (stderr, "%s:  unable to find a username for you.\n",
		     "Mailbox widget");
	    CloseDown (w, 1);
	}
	username = pw->pw_name;
    }
    w->ssmailbox.filename = (String) XtMalloc (strlen (MAILBOX_DIRECTORY) + 1 +
				   	     strlen (username) + 1);
    strcpy (w->ssmailbox.filename, MAILBOX_DIRECTORY);
    strcat (w->ssmailbox.filename, "/");
    strcat (w->ssmailbox.filename, username);
    return;
}

static void CloseDown (w, status)
    SSMailboxWidget w;
    int status;
{
    Display *dpy = XtDisplay (w);

    XtDestroyWidget (w);
    XCloseDisplay (dpy);
    exit (status);
}


/* ARGSUSED */
static Boolean SetValues (gcurrent, grequest, gnew)
    Widget gcurrent, grequest, gnew;
{
    SSMailboxWidget current = (SSMailboxWidget) gcurrent;
    SSMailboxWidget new = (SSMailboxWidget) gnew;
    Boolean redisplay = FALSE;

    if (current->ssmailbox.update != new->ssmailbox.update) {
	if (current->ssmailbox.interval_id) 
	  XtRemoveTimeOut (current->ssmailbox.interval_id);
	new->ssmailbox.interval_id = XtAddTimeOut (new->ssmailbox.update * 1000,
						 clock_tic,
						 (caddr_t) gnew);
    }

    if (current->ssmailbox.foreground_pixel != new->ssmailbox.foreground_pixel ||
	current->core.background_pixel != new->core.background_pixel) {
	XtDestroyGC (current->ssmailbox.gc);
	new->ssmailbox.gc = get_ssmailbox_gc (new);
	redisplay = TRUE;
    }

    return (redisplay);
}


/*
 * drawing code
 */

static void redraw_ssmailbox (w)
    SSMailboxWidget w;
{
    register Display *dpy = XtDisplay (w);
    register Window win = XtWindow (w);
    register int x, y;
    GC gc = w->ssmailbox.gc;
    Pixel back = w->core.background_pixel;
    struct _mbimage *im;

    /* center the picture in the window */

    if (w->ssmailbox.flag_up) {		/* paint the "up" position */
	im = &w->ssmailbox.full;
	if (w->ssmailbox.flipit) back = w->ssmailbox.foreground_pixel;
    } else {				/* paint the "down" position */
	im = &w->ssmailbox.empty;
    }
    x = (((int)w->core.width) - im->width) / 2;
    y = (((int)w->core.height) - im->height) / 2;

    XSetWindowBackground (dpy, win, back);
    XClearWindow (dpy, win);
    XCopyArea (dpy, im->pixmap, win, gc, 0, 0, im->width, im->height, x, y);

#ifdef SHAPE
    /*
     * XXX - temporary hack; walk up widget tree to find top most parent (which
     * will be a shell) and mash it to have our shape.  This will be replaced
     * by a special shell widget.
     */
    if (w->ssmailbox.shapeit) {
	Widget parent;

	for (parent = (Widget) w; XtParent(parent);
	     parent = XtParent(parent)) {
	    x += parent->core.x + parent->core.border_width;
	    y += parent->core.y + parent->core.border_width;
	}

	if (im->mask != w->ssmailbox.shape_cache.mask ||
	    x != w->ssmailbox.shape_cache.x || y != w->ssmailbox.shape_cache.y) {
	    XShapeCombineMask (XtDisplay(parent), XtWindow(parent),
			       ShapeBounding, x, y, im->mask, ShapeSet);
	    w->ssmailbox.shape_cache.mask = im->mask;
	    w->ssmailbox.shape_cache.x = x;
	    w->ssmailbox.shape_cache.y = y;
	}
    }
#endif

    return;
}

#ifdef sparc

/*
 * WorkProc to play sound in idle time
 */
static Boolean
playOneWP(w)
SSMailboxWidget w;
{
    if (play_sound(w->ssmailbox.sound, w->ssmailbox.volume) < 0)
	XBell(XtDisplay(w), w->ssmailbox.volume);

    return(True);
}

static Boolean
playManyWP(soundp)
struct PlaySound *soundp;
{
    struct SoundStruct *ss;
    struct MsgStruct *sounds;

    sounds = soundp->msg;
    while (sounds && soundp->num_sounds) 
    {
	ss = sounds->sounds;
	while(ss) 
	{
	    if (play_sound(ss->sound, soundp->volume) < 0)
		XBell(soundp->dpy, soundp->volume);
	    ss = ss->next;
	}
	sounds = sounds->next;
	--soundp->num_sounds;
    }

    free_sounds(soundp->msg);

    XtFree(soundp);
    return(True);
}
#endif

static void beep (w, sounds, num)
    SSMailboxWidget w;
    struct MsgStruct *sounds;
    int num;
{
#ifdef sparc
    struct PlaySound *play;
    
    if (sounds) {
	play = XtNew(struct PlaySound);
	play->dpy = XtDisplay(w);
	play->num_sounds = num;
	play->volume = w->ssmailbox.volume;
	play->msg = sounds;

	(void) XtAddWorkProc(playManyWP, play);
    }
    else if (w->ssmailbox.sound) {
	play = XtNew(struct PlaySound);
	play->dpy = XtDisplay(w);
	play->volume = w->ssmailbox.volume;
	play->sound = w->ssmailbox.sound;

	(void) XtAddWorkProc(playOneWP, play);
    }
    else
#endif
	XBell (XtDisplay (w), w->ssmailbox.volume);
    return;
}
#define LSIZE 256

void add_sound();

static int newmsgs(w, sounds)
SSMailboxWidget w;
struct MsgStruct **sounds;
{
   char *mbox;
   FILE *fp;
   int mcnt, status, numnew, in_header;
   static char line[LSIZE];
   static char save_line[LSIZE];
   static char tmp_line[LSIZE];
   int save_flag = 0;
   Bindings *binding;
   struct MsgStruct *ms = 0;

   mbox = w->ssmailbox.filename;
    
   /* if we can't open the mailbox return 0.
    */
   if (!(fp= fopen(mbox, "r"))) {
	   return 0;
   }
    
   /* read in the sender and subject of each letter
    */
   mcnt = -1; status = numnew = 0; in_header=1;
    
   while ( fgets(line, LSIZE, fp) ) {
      if (!strncmp(line, "From ", 5)) {
         if(save_flag == 1) {
            sprintf(tmp_line, "From: %s", &save_line[5]);
            for(binding=w->ssmailbox.bindings; binding;
                binding = binding->next) {
               if (regexec(binding->pattern, tmp_line)) {
                  struct SoundStruct *s;
                  if (!ms) {
                     ms = (struct MsgStruct *)malloc(sizeof(struct MsgStruct));
                     ms->sounds = 0;
                     ms->next = 0;
                  }
                  s = (struct SoundStruct *)malloc(sizeof(struct SoundStruct));
                  s->sound = binding->sound;
                  s->next = ms->sounds;
                  ms->sounds = s;
               }
            }
         }

         if (!status && mcnt != -1) {
            numnew++;
            add_sound(w, sounds, ms, status);
         }
         ms = 0;
         mcnt++;
         status = 0;
         in_header = 1;

         strcpy(save_line, line);
         save_flag = 1;

      } else if (in_header && *line == '\n') {
         in_header = 0;
      } else if (in_header && (!strncmp(line, "From: ", 6) ||
			       !strncmp(line, "Cc: ", 4) ||
			       !strncmp(line, "Subject: ", 9) ||
			       !strncmp(line, "To: ", 4))) {
         for(binding=w->ssmailbox.bindings; binding; binding = binding->next) {
            if (regexec(binding->pattern, line)) {
               struct SoundStruct *s;
               if (!ms) {
                  ms = (struct MsgStruct *)malloc(sizeof(struct MsgStruct));
                  ms->sounds = 0;
                  ms->next = 0;
               }
               s = (struct SoundStruct *)malloc(sizeof(struct SoundStruct));
               s->sound = binding->sound;
               s->next = ms->sounds;
               ms->sounds = s;
               save_flag = 0;
            }
         }
      } else if (!status && in_header && !strncmp(line, "Status: ", 8)) {
         status = 1;
      }
   }

   if(save_flag == 1) {
      sprintf(tmp_line, "From: %s", &save_line[5]);
      for(binding=w->ssmailbox.bindings; binding;
          binding = binding->next) {
         if (regexec(binding->pattern, tmp_line)) {
            struct SoundStruct *s;
            if (!ms) {
               ms = (struct MsgStruct *)malloc(sizeof(struct MsgStruct));
               ms->sounds = 0;
               ms->next = 0;
            }
            s = (struct SoundStruct *)malloc(sizeof(struct SoundStruct));
            s->sound = binding->sound;
            s->next = ms->sounds;
            ms->sounds = s;
         }
      }
   }

   if (!status && mcnt != -1) {
	   numnew++;
      add_sound(w, sounds, ms, status);
   }

   fclose(fp);
   return numnew;
}

static void add_sound(w, sounds, ms, status)
    SSMailboxWidget w;
    struct MsgStruct **sounds, *ms;
    int status;
{
    if (!status) 
    {
	if (!ms && w->ssmailbox.sound)
	{
	    ms = (struct MsgStruct *)malloc(sizeof(struct
						   MsgStruct));
	    ms->sounds = (struct SoundStruct *)malloc(sizeof(struct
							     SoundStruct));
	    ms->sounds->sound = w->ssmailbox.sound;
	    ms->sounds->next = 0;
	}
	if (ms) {
	    ms->next = *sounds;
	    *sounds = ms;
	}
    }
    else if (ms)
    {
	while (ms->sounds) 
	{
	    struct SoundStruct *x;
	    x = ms->sounds->next;
	    free((char *)ms->sounds);
	    ms->sounds = x;
	}
	free((char *)ms);
    }
}

void regerror(s)
    char *s;
{
    fprintf(stderr, "regexp(3): %s\n", s);
}
