#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/signal.h>

#include <stropts.h>
#include <sys/ioctl.h>

#include <multimedia/libaudio.h>
#include <multimedia/audio_device.h>

extern char		*ProgramName;

#define	MAX_GAIN		(100)		/* maximum gain */

/*
 * This defines the tolerable sample rate error as a ratio between the
 * sample rates of the audio data and the audio device.
 */
#define	SAMPLE_RATE_THRESHOLD	(.01)

/*
 * Play an audio file.
 */
int
play_sound(sound_file, volume)
char *sound_file;
int volume;
{
	double	Savevol;				/* saved volume level */
	int		Audio_fd = -1;			/* file descriptor for audio device */
	Audio_hdr	Dev_hdr;			/* audio header for device */
	Audio_hdr	File_hdr;			/* audio header for file */
	unsigned char	buf[1024 * 64];	/* size should depend on sample_rate */
	struct sigvec vec;
	int		cnt;
	int		err;
	int		ifd;
	double		vol;
	struct stat	st;

	/* Validate and open the audio device */
	err = stat("/dev/audio", &st);
	if (err < 0) {
		fprintf(stderr, "%s: cannot stat /dev/audio\n", ProgramName);
		return(-1);
	}
	if (!S_ISCHR(st.st_mode)) {
		fprintf(stderr, "%s: /dev/audio is not an audio device\n",
			ProgramName);
		return(-1);
	}

	/* Try it quickly, first */
	Audio_fd = open("/dev/audio", O_WRONLY | O_NDELAY);
	if ((Audio_fd < 0) && (errno == EBUSY)) {
		fprintf(stderr, "%s: /dev/audio is busy\n", ProgramName);
		return(-1);
	}
	if (Audio_fd < 0) {
		fprintf(stderr, "%s: error opening /dev/audio\n", ProgramName);
		return(-1);
	}

	/* Get the device output encoding configuration */
	if (audio_get_play_config(Audio_fd, &Dev_hdr) != AUDIO_SUCCESS) {
		fprintf(stderr, "%s: /dev/audio is not an audio device\n",
		    ProgramName);
		close(Audio_fd);
		return(-1);
	}

	/* set the volume */
	vol = (double) volume / (double) MAX_GAIN;
	(void) audio_get_play_gain(Audio_fd, &Savevol);
	err = audio_set_play_gain(Audio_fd, &vol);
	if (err != AUDIO_SUCCESS) {
		fprintf(stderr,
		    "%s: could not set output volume for /dev/audio\n",
		    ProgramName);
		close(Audio_fd);
		return(-1);
	}

	/* if we don't set a signal handler, the volume we set above
	   is ignored??!! so set up the default handler */
	vec.sv_handler = SIG_DFL;
	vec.sv_mask = 0;
	vec.sv_flags = 0;
	(void) sigvec(SIGINT, &vec, (struct sigvec *)NULL);

	if ((ifd = open(sound_file, O_RDONLY, 0)) < 0) {
		fprintf(stderr, "%s: cannot open %s\n", ProgramName,
			sound_file);
		close(Audio_fd);
		return(-1);
	}

	/* Check to make sure this is an audio file */
	err = audio_read_filehdr(ifd, &File_hdr, (char *)NULL, 0);

	if (err != AUDIO_SUCCESS) {
		fprintf(stderr, "%s: %s is not a valid audio file\n",
		    ProgramName, sound_file);
		close(Audio_fd);
		close(ifd);
		return(-1);
	}

	/* Check the device configuration */
	if (audio_cmp_hdr(&Dev_hdr, &File_hdr) != 0) {
		close(Audio_fd);
		close(ifd);
		return(-1);
	}

	/*
	 * At this point, we're all ready to copy the data.
	 */
	while ((cnt = read(ifd, (char *)buf, sizeof (buf))) >= 0) {
		/* If input EOF, write an eof marker */
		err = write(Audio_fd, (char *)buf, cnt);

		if (err != cnt)
			break;
		if (cnt == 0)
			break;
	}
	if (cnt < 0 || err != cnt) {
		fprintf(stderr, "%s: error reading %s\n",
			ProgramName, sound_file);
		close(Audio_fd);
		close(ifd);
		return(-1);
	}

	(void) close(ifd);		/* close input file */

	/* reset volume */
	(void) audio_drain(Audio_fd, FALSE);
	(void) audio_set_play_gain(Audio_fd, &Savevol);
	
	(void) close(Audio_fd);		/* close output */
	return(0);
}
